(* contains compiler for ppl language *)
(* $Id: compiler.ml,v 1.50 2000/06/27 19:28:54 alamstei Exp $ *)

open Pseudo
open Library
open Lambda

type symbol_table = (Lambda.variable * value) list

(* Following two functions set a default warning and then make a reference to 
   it (and so make it mutable).  Useful to alert user that some code may 
   contain errors.  *)
      
let default_warning s = 
  prerr_endline ("Warning: "^ s)
    
let warning = ref default_warning
  
(* find what the variable var maps to in the symbol-table table *)
let table_lookup var table = 
  try
    List.assoc var table
  with Not_found ->
    raise (Exn.Unbound_variable var)
      
(* return a pair of lists (free,b) b is a list of consecutive memory locations
   starting from Mem(c,i) *)
let rec generate_table' c i free = 
  match free with
    (hd::tl) ->
      (hd,Lv(Mem(c,i)))::(generate_table' c (i+1) tl)
  | [] -> []
	
(* switch v n [c(0);...;c(k-1)] generates code which checks the value v and 
   then executes the continuation c(v-n) *)
let rec switch v n continuations =
  match continuations with
    hd::[] ->
      (hd,[])
  | hd::tl ->
      let label = new_label() in
      let (remaining_code,remaining_segments) = switch v (n+1) tl in
      let code_portion = (CMP (v,N n, 
			       BE (Label label, remaining_code))) in
      let this_segment = [(label,hd)] in
      let segment_portion = this_segment @ remaining_segments in
      (code_portion,segment_portion)
  | [] -> raise (Exn.Compiler_exception "Compiler.switch")

(* save_values valuelist a i cont: makes code that stores the Pseudo.values 
   from the valuelist at consecutive memory locations starting from Mme(a,i), 
   and then executes cont. *)
let rec save_values valuelist a i cont =
  match valuelist with
    [] -> cont
  | (hd::tl) -> LOAD (Mem(a,i), hd, save_values tl a (i+1) cont)

(* terms_of_injlist [inj1;...;injn]:  Takes a list of injections and returns 
   them in the appropriate order as lambda terms.  For example, each injection
   of the form (i,j,x,M) is mapped to (\x.M).  A warning is issued if the list
   contains duplicate entries, and only the first instance is mapped.  A 
   warning is also issued if a case is missing. *)

let terms_of_injlist injlist =

  (* an auxiliary function that extracts all j/n injections from the injlist *)
  let rec extract_jth j n injlist =
    match injlist with
      []                               -> []
    | (j',n',x,m)::t when j=j' && n=n' -> Abs(x,m)::extract_jth j n t
    | h::t                             -> extract_jth j n t
  in

  (* an auxiliary function that extracts *the* j/n injection from the injlist,
     and gives appropriate warning if there is more or less than one *)
  let extract_the j n injlist =
    match extract_jth j n injlist with
      [] ->
      	(!warning) ("missing case "^ Lambda.string_of_inj j n ^ 
	  " in case distinction"); 
	Var (Uservar "match#failure")
    | [x] -> x
    | x::y::t ->
      	(!warning) ("multiple case "^ Lambda.string_of_inj j n ^ 
          " in case distinction");
	x
  in
  
  (* first determine n *)
  match injlist with
    [] -> []
  | (_,n,_,_)::_ ->
      (* build the list ns = [1;...;n] *)
      let ns = List.map (fun x->x+1) (Library.create n) in
      (* build the list of terms *)
      List.map (fun j -> extract_the j n injlist) ns     

(* build_multiclosure [m1;...;mk] table a cont: generates assembly code which 
   builds a multiclosure for terms [m1;...;mk], puts a poitner to it in the 
   register a, and then executes cont.  Returns a code-segments pair.  A 
   multiclosure is the following: suppose m1...mk contain the free variables 
   y1...yn.  Then the multiclosure consists of k+n consecutive words of  
   memory.  The first k words contain pointers to the codes for m1...mk, and 
   the next n words contain the values of the free variables y1...yn.  "to 
   invoke the jth branch of the multiclosure" means to put a pointer to the 
   multiclosure in the register C and then to jump to the code for mj.  A 
   regular closure is the special case k=1, and a match closure is the special
   case where each mj is of the form \x.n. *)

let rec build_multiclosure terms table a cont =
  let labels = List.map (fun x -> new_label()) terms in
  let labelvalues = List.map (fun x -> Label x) labels in
  let nterms = List.length terms in

  let free = Library.set_flatten (List.map free_vars_of terms) in
  let freevalues = List.map (fun x -> table_lookup x table) free in
  let nfree = List.length free in

  (* first build the code portion *)
  let code_portion = (ALLOC (Reg a, N(nterms + nfree),
			     save_values (labelvalues @ freevalues) a 0 cont))
  in
    
  (* now build the segment portion *)
  let table' = generate_table' "C" nterms free in
  let segs_of_term t l =
    let (code_for_t, segs_for_t) = code_for t table' in
    (l, code_for_t) :: segs_for_t
  in
  (code_portion, List.flatten (List.map2 segs_of_term terms labels))

and build_closure term table a cont =
  build_multiclosure [term] table a cont

and build_match_closure injlist table a cont =
  build_multiclosure (terms_of_injlist injlist) table a cont

(* generate assembly code for [[term]]_table *)
and code_for term table = 
  match term with
    Var(x) -> 
      let sx = (table_lookup x table) in
      let c = "C" in
      let code_portion = (LOAD (Reg c, sx, JUMP (Lv(Mem(c,0))))) in
      (code_portion, [])
  | Abs(x,m) ->
      let regx = Reg(new_register()) in
      let new_table = (x,Lv(regx))::table in
      let (code_for_m, segments_for_m) = code_for m new_table in
      let code_portion = (EXITOES ("fn",POP (regx, code_for_m))) in
      (code_portion, segments_for_m)
  | App(m,Var(x)) ->
      let (code_for_m, segments_for_m) = code_for m table in
      let code_portion = (PUSH ((table_lookup x table), code_for_m)) in
      (code_portion, segments_for_m)
  | App(m,n) -> 
      let (code_for_m, segs_for_m) = code_for m table in
      let cont = PUSH(Lv (Reg "A"), code_for_m) in
      let (code, segs) = build_closure n table "A" cont in 
      (code, segs @ segs_for_m)
  | Let(x,m,n) ->
      (* treat Let(x,m,n) as App(Abs(x,n),m), but with stream-lining *)
      let regx = new_register() in 
      let new_table = (x,Lv(Reg regx))::table in
      let (code_for_n, segs_for_n) = code_for n new_table in
      let (code, segs) = build_closure m table regx code_for_n in
      (code, segs @ segs_for_n)
  | Letrec(x,m,n) ->
      let regx = new_register() in 
      let new_table = (x,Lv(Reg regx))::table in
      let (code_for_n, segs_for_n) = code_for n new_table in
      let (code, segs) = build_closure m new_table regx code_for_n in
      (code, segs @ segs_for_n)
  | Intlit(n) ->
      let v = "V" in
      let r = "R" in
      let regv = Reg(v) in
      let regr = Reg(r) in
      let code_portion = (LOAD(regv, N(n), EXITOESINT (Lv(regv), 
			  POP (regr, JUMP (Lv(regr)))))) in
      (code_portion, [])
  | Boollit(b) ->
      let n = (if b then 1 else 0) in
      let v = "V" in
      let r = "R" in
      let regv = Reg(v) in
      let regr = Reg(r) in
      let code_portion = (LOAD(regv, N(n), EXITOESBOOL (Lv(regv), 
			  POP (regr, JUMP (Lv(regr)))))) in
      (code_portion, [])
  | Proj(i,j,m) ->
      let (code_for_m,segments_for_m) = code_for m table in
      let code_portion = PUSH(N(i),code_for_m) in
      (code_portion,segments_for_m)
  | Tuple(tuplelist) ->
      let len = List.length tuplelist in
      if (len=0) then 
        (EXIT "()", [])
      else
        let strlen = string_of_int len in
        let newreg = new_register() in
      	let newreg' = Reg(newreg) in
      	let cf = fun tuples ->  code_for tuples table in
      	let continuations = List.map cf tuplelist in
      	let (codes_for_conts,segs_for_conts) = List.split continuations in
      	let segs' = List.flatten segs_for_conts in
      	let (code_for_switch,segs_for_switch) = 
	  switch (Lv newreg') 1 codes_for_conts in
      	(EXITOES ((strlen ^"-tuple"), POP (newreg', code_for_switch)),
	 (segs_for_switch @ segs'))
  | Inj(i,j,m) -> 
      let cont = PUSH(Lv (Reg "A"), JUMP(Lv (Mem("C", i-1)))) in
      let (code_body, segs) = build_closure m table "A" cont in
      let strinj = string_of_inj i j in
      let code = EXITOES(strinj, POP(Reg "C", code_body)) in
      (code, segs)
  | Case(m,injlist) ->
      let (code_for_m, segs_for_m) = code_for m table in
      let cont = PUSH(Lv (Reg "A"), code_for_m) in
      let (code, segs_for_match) = 
	build_match_closure injlist table "A" cont in
      (code, segs_for_m @ segs_for_match)
	  
let compile term =
  code_for term Stdlib.stdbindings
